/*
 *  TOPPERS/JSP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Just Standard Profile Kernel
 * 
 *  Copyright (C) 2000,2001 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 * 
 *  嵭ԤϡFree Software Foundation ˤäƸɽƤ 
 *  GNU General Public License  Version 2 ˵ҤƤ狼
 *  ξΤ줫˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥѲǽʥХʥꥳɡʥ֥륪
 *      ȥե饤֥ʤɡˤηѤˤϡ
 *      ȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭ɽ
 *      Ѿ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥԲǽʥХʥꥳɤηޤϵ
 *      ߹ѤˤϡΤ줫ξȡ
 *    (a) ѤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) Ѥη֤̤ˡˤäơ嵭Ԥ𤹤
 *        ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭Ԥդ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥϡ
 *  ܥեȥ˴ؤơŬѲǽޤơʤݾڤԤ
 *  ʤޤܥեȥѤˤľŪޤϴŪ
 *  ʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: fc_bfd.cpp,v 1.3 2001/11/12 14:59:27 takayuki Exp $
 */


#include "filecontainer.h"

#include "exception.h"
#include "message.h"

#include <bfd.h>
#include <libiberty.h>

#include <string>
#include <map>

using namespace std;


/*
 *  BFDȤäե륳ƥʥ饹
 */
class filecontainer_BFD : public filecontainer
{
protected:
	bfd * object;							//֥
	map<string, asymbol *> symbol_table;	//ѥϥåɤ
	asymbol ** symbol_container;

public:
	filecontainer_BFD(void);
	~filecontainer_BFD(void);

	virtual bool attach_module(const char *);
	virtual bool change_endian(void * target, unsigned int size);
	virtual bool load_contents(void * dest, unsigned long address, unsigned int size);
	virtual unsigned long get_symbol_address(const char *);
	virtual const char * get_architecture(void);

} Instance_of_filecontainer_BFD;

filecontainer_BFD::filecontainer_BFD(void)
{
	bfd_init();

	object = NULL;
	symbol_table.clear();
	symbol_container = NULL;
	container = this;
}

filecontainer_BFD::~filecontainer_BFD(void)
{
	if(object != NULL)
		bfd_close(object);
}

/*
 *  attach_module : оݥ⥸塼򥢥å
 */
bool filecontainer_BFD::attach_module(const char * filename)
{
	char ** target_list;
	asymbol ** symbols;
	asymbol *  sym;
	int num_syms;
	boolean result;
	int i;

	if(object != NULL)
		bfd_close(object);
	symbol_table.clear();

		//⥸塼륪ץ (ɹ)
	object = bfd_openr(filename, "default");

		//åȲ
	target_list = (char **)bfd_target_list();
	result = bfd_check_format_matches(object, bfd_object, &target_list);
	if(result == 0)
	{
		Exception::Throw(Message("Internel error: BFD could not recognize the target file format.","顼: BFDϥեɤ߽Ф˼Ԥޤ"));
		return false;
	}

		//ܥΥϥåɤ
	symbols = (asymbol **)xmalloc( bfd_get_symtab_upper_bound(object) );
	num_syms = bfd_canonicalize_symtab(object, symbols);

	for(i=0;i<num_syms;i++)
	{
		sym = *(symbols+i);
		if(sym != NULL && sym->name != NULL && *(sym->name) != '\x0')
			symbol_table[string(sym->name)] = sym;
	}

	symbol_container = symbols;
	return true;
}

bool filecontainer_BFD::change_endian(void * target, unsigned int size)
{
		//ۥȤϥȥȲ
	enum bfd_endian host_endian = BFD_ENDIAN_LITTLE;

	char * top, * tail;

	if(object == NULL)
		return false;

	if(object->xvec->byteorder == BFD_ENDIAN_UNKNOWN)
		return false;

/*
	unsigned int __work = 0x1;
	if( *(char *)__work == 0)
		host_endian = BFD_ENDIAN_BIG;
*/

	if(object->xvec->byteorder == host_endian)
		return true;

	/*
	 *  ᥤ롼 : bswapʤǡľ˽
	 */

	top  = (char *)target;
	tail = (char *)target+size -1;

	while(top < tail)
	{
		*top ^= *tail, *tail ^= *top, *top ^= *tail;
		top ++;
		tail --;
	}

	return true;
}


bool filecontainer_BFD::load_contents(void * dest, unsigned long address, unsigned int size)
{
	struct sec * section;

	if(object == 0)
		return false;

		//оݥɥ쥹ͭ륻õ
	section = object->sections;
	while(section != 0)
	{
		if(address - (unsigned long)section->vma <= section->_raw_size 
			&& (section->flags & (SEC_ALLOC|SEC_HAS_CONTENTS)) == (SEC_ALLOC|SEC_HAS_CONTENTS))
		{
				//ɤ߽Ф
			bfd_get_section_contents(object, section, dest, address - (unsigned long)section->vma, size);
			return true;
		}
		section = section->next;
	}

		//ɤˤʤ
	Exception::Throw(Message("Internel error: Memory read with unmapped address","顼; ޥåפƤʤɥ쥹Ȥäƥ꡼ɤԤޤ"));

	return false;
}


/*
 * get_symbol_address : ܥ̾饢ɥ쥹ͤ
 */
unsigned long filecontainer_BFD::get_symbol_address(const char * symbol)
{
	map<string, asymbol *>::iterator scope;
	string symbol_name;

	if(object == 0)
	{
		Exception::Throw(Message("Not initialized","Ƥޤ"));
		return 0;
	}

		//ܥ̾ ("_"Ȥν)
	if(object->xvec->symbol_leading_char != '\x0')
		symbol_name += object->xvec->symbol_leading_char;
	symbol_name += symbol;

	scope = symbol_table.find(symbol_name);
	if(scope == symbol_table.end())
	{
		Exception::Throw(string(Message("Internal error: Unknown symbol [","顼: ʥܥ [")) + symbol_name + "]");
		return 0;
	}

		//Address = ⥪եå + VMA
	return (*scope).second->value + (*scope).second->section->vma;
}

/*
 * get_architecture : ƥ̾μ
 */
const char * filecontainer_BFD::get_architecture(void)
{
	if(object == NULL)
		return "Unknown";

		//Ȥꤢå̾äƥƥ̾ˤƤ
		// #ɤХʡˤȤäƤʤ ΤȤ
	return object->xvec->name;
}
