/*
 *  TOPPERS/JSP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Just Standard Profile Kernel
 * 
 *  Copyright (C) 2000,2001 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 * 
 *  嵭ԤϡFree Software Foundation ˤäƸɽƤ 
 *  GNU General Public License  Version 2 ˵ҤƤ狼
 *  ξΤ줫˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥѲǽʥХʥꥳɡʥ֥륪
 *      ȥե饤֥ʤɡˤηѤˤϡ
 *      ȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭ɽ
 *      Ѿ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥԲǽʥХʥꥳɤηޤϵ
 *      ߹ѤˤϡΤ줫ξȡ
 *    (a) ѤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) Ѥη֤̤ˡˤäơ嵭Ԥ𤹤
 *        ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭Ԥդ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥϡ
 *  ܥեȥ˴ؤơŬѲǽޤơʤݾڤԤ
 *  ʤޤܥեȥѤˤľŪޤϴŪ
 *  ʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: parser.h,v 1.5 2001/11/12 14:59:27 takayuki Exp $
 */


#ifndef __PARSER_H
#define __PARSER_H

#include <iostream>
#include <list>
#include <string>

class Token : public std::string
{
public:
	enum tagTokenType
	{
		IDENTIFIER		= 0x01,
		INTEGER			= 0x02,
		STRINGLITERAL	= 0x03,
		OPERATOR		= 0x04,
		PUNCTUATOR		= 0x05,
		RESERVEDWORD	= 0x06,
		SPECIAL			= 0x80,
		SPACE			= 0x81,
		UNKNOWN			= 0xfe,
		ERROR			= 0xff
	};

	enum tagTokenType Type;
	long Value;

	Token(void)			{ Type = UNKNOWN;  Value = 0;	};
	Token(enum tagTokenType, const char *);
	Token(const Token & src)	{ (*this) = src; };

	operator const enum tagTokenType (void) const { return Type; };

	Token & operator =(const Token & src);
	bool operator == (const Token & src) const;

	Token & Trim(void);
};

class Parser
{
protected:
	std::istream * in;
	static const char * Punctuator;
	static const char * Operator;
	std::list<Token> TokenStack;

	bool GetIdentifier(Token &, int);
	bool GetWhitespace(Token &, int);
	bool GetInteger(Token &, int);
	bool GetStringLiteral(Token &, int);
	bool GetOperator(Token &, int);

public:
	Parser(void);
	~Parser(void) {};

	std::istream * SetStream(std::istream *);
	enum Token::tagTokenType GetToken(Token &,bool = false);
	void GetToken(Token &,enum Token::tagTokenType, const char * = NULL);
	void PutBack(Token &);
};

class ParserComponent
{
public:
	virtual void Parse(class Manager *, Parser *) =0;
};

inline std::istream * Parser::SetStream(std::istream * _in)
{
	std::istream * old = in;
	in = _in;
	return old;
}

inline Token::Token(enum tagTokenType type, const char * term)
{
	Type = type;
	Value = 0;
	assign(term);
}

inline Token & Token::operator =(const Token & src)
{	
	Type = src.Type; 
	Value = src.Value; 
	assign(src); 
	return *this;	
}

inline bool Token::operator ==(const Token & src) const
{
	if(Type != src.Type)
		return false;
	if(Type == Token::INTEGER && Value != src.Value)
		return false;
	else
		if(compare(src) != 0)
			return false;

	return true;
}

inline void Parser::PutBack(Token & token)
{	TokenStack.push_front(token);	}

#endif