/*
 *  TOPPERS/JSP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Just Standard Profile Kernel
 * 
 *  Copyright (C) 2000,2001 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 * 
 *  嵭ԤϡFree Software Foundation ˤäƸɽƤ 
 *  GNU General Public License  Version 2 ˵ҤƤ狼
 *  ξΤ줫˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥѲǽʥХʥꥳɡʥ֥륪
 *      ȥե饤֥ʤɡˤηѤˤϡ
 *      ȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭ɽ
 *      Ѿ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥԲǽʥХʥꥳɤηޤϵ
 *      ߹ѤˤϡΤ줫ξȡ
 *    (a) ѤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) Ѥη֤̤ˡˤäơ嵭Ԥ𤹤
 *        ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭Ԥդ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥϡ
 *  ܥեȥ˴ؤơŬѲǽޤơʤݾڤԤ
 *  ʤޤܥեȥѤˤľŪޤϴŪ
 *  ʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: value.h,v 1.2 2001/11/12 14:59:27 takayuki Exp $
 */


#ifndef __VALUE_H
#define __VALUE_H

#include <vector>
#include <string>
#include <iostream>

class Value
{
public:
	enum tagType
	{
		UNKNOWN = 0,
		INTEGER,
		STRING,
		FLOAT,
		ARRAY
	};

	Value(void) {}
	virtual ~Value(void){};

	virtual operator enum tagType (void) const = 0;

	virtual Value * Duplicate(void) = 0;
	virtual void Serialize(std::ostream &) const = 0;
};

class Integer : public Value
{
protected:
	int value;

public:
	Integer(long src=0) { value = src; };
	Integer(const Integer & src) { value = src.value; };
	virtual ~Integer(void) {};

	virtual operator enum tagType (void) const { return INTEGER; };
	
	operator const int (void) const;
	operator int & (void);

	virtual Value * Duplicate(void);
	virtual void Serialize(std::ostream &) const;
};

class String : public Value, public std::string
{
public:
	String(void) {};
	String(const char *);
	String(std::string &);
	String(const String & src);
	virtual ~String(void) {};
	virtual operator enum tagType (void) const { return STRING; };
	virtual Value * Duplicate(void);
	virtual void Serialize(std::ostream &) const;
};

class Array : public Value, public std::vector<Value *>
{
public:
	Array(void) {};
	Array(const Array &);
	virtual ~Array(void);

	virtual operator enum tagType (void) const { return ARRAY; };
	Array & operator =(const Array &);

	void clear();
	
	virtual Value * Duplicate(void);
	virtual void Serialize(std::ostream &) const;
};


//--- Inline functions

inline Integer::operator const int (void) const
{	return value;	}

inline Integer::operator int & (void)
{	return value;	}

inline Value * Integer::Duplicate(void)
{	return new Integer(value);	}

inline void Integer::Serialize(std::ostream & out) const
{	out << value;	}

inline String::String(const char * str)
{	assign(str);	}

inline String::String(std::string & str)
{	assign(str.c_str());	}

inline String::String(const String & src)
{	assign(src);	}

inline Value * String::Duplicate(void)
{	return static_cast<Value *>(new String(*this));	}

inline void String::Serialize(std::ostream & out) const
{	out << static_cast<std::string>(*this);	}


inline Value * Array::Duplicate(void)
{	return new Array(*this);	}

//--- Related functions

inline std::ostream & operator <<(std::ostream & out, Value & val)
{
	val.Serialize(out);
	return out;
}

inline std::ostream & operator <<(std::ostream & out, Value * val)
{
	if(val != NULL)
		val->Serialize(out);
	else
		out << "<null>";
	return out;
}

#endif